/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Properties;

import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanServerConnection;
import javax.management.ObjectName;
import javax.management.RuntimeMBeanException;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;
import javax.swing.JOptionPane;

/**
 * This is a very basic worker launcher. It reads the management properties of
 * the home process, figures out the local port and connects via JMX.
 * With a little work we could do much smarter things. However... this is simple
 * basic, and it works.
 * <p>
 * See also:
 * <ul>
 * <li>{@link http://blogs.sun.com/jmxetc/entry/how_to_retrieve_remote_jvm}</li>
 * <li>{@link http://java.sun.com/j2se/1.5.0/docs/guide/management/agent.html}</lI>
 * </ul> 
 * 
 * @author hb
 *
 */

public class LaunchWorker {

	private static final String GUI = "gui";
	private static final File MANAGEMENT_PROPERTIES = new File("management.properties");
	private static final String JMX_PORT = "com.sun.management.jmxremote.port";

	private String workerName;
	private boolean gui;
	
	private boolean checkArgs(String[] args) {
		for (int i = 0; i < args.length;) {
			String option = args[i++];
			if (option.startsWith("-")) {
				option = option.substring(1);
				if (option.equals("?")) {
					usage();
					return false;
				} else if (option.equals(GUI)) {
					this.gui=true;
					continue;
				}
			} else 
			if (this.workerName==null) {
				this.workerName = option;
				continue;
			}
			_info("SEVERE: unknown option or duplicate worker specification: " + option);
			usage();
			return false;
		} 
		if (this.workerName==null) {
			_error("Please specify a worker to start.", null);
			return false;
		}
		return true;
	}


	private void run(String[] args) {
		hello();
		if (!checkArgs(args)) {
			usage();
		}

		Properties p = new Properties();
		FileInputStream fin=null;
		try {
			try { 
				fin = new FileInputStream(MANAGEMENT_PROPERTIES);
				p.load(fin); 
			} catch (FileNotFoundException fnfe) {
				_error("Failed to find management.properties.\nNOTE: The worker launcher must be executed in the \"bin\" folder of your home installation.",null);
				return;
			} finally {
				if (fin!=null)
					fin.close();
			}
		} catch (IOException ioe) {
			_error("Failed to read management.properties.",ioe);
			return;
		}
		/*
		 * BIG TODO: make this secure. For example, let the home process generate a ticket... 
		 */
		String ps = p.getProperty(JMX_PORT);
		if (ps==null) {
			_error("Could not find JMX port ("+JMX_PORT+") in management.properties.\nNOTE: The worker launcher requires a JMX port to connect to the home process.",null);
			return;
		}
		// ok, now connect
		MBeanServerConnection mbs = null;
		try {
			JMXServiceURL url = new JMXServiceURL("service:jmx:rmi:///jndi/rmi://localhost:" + ps +  "/jmxrmi");
			JMXConnector  con = JMXConnectorFactory.connect(url);
			mbs = con.getMBeanServerConnection();
		} catch (IOException e) {
			_error("Failed to connect to home process. \nIs a home process running at jmx port "+ps+"?",null);
			return;
		} 

		// get the workers mbean
		ObjectName on =null;
		try {
			on = ObjectName.getInstance("zfabrik:type=com.zfabrik.impl.workers.home.HomeLayoutResource");
			mbs.invoke(on,"startWorker", new Object[]{this.workerName}, new String[]{String.class.getName()});
		} catch (InstanceNotFoundException e) {
			_error("Failed to retrieve workers MBean ("+on+")",e);
		} catch (MBeanException e) {
			_error("Error during MBean invocation.",e.getCause());
		} catch (RuntimeMBeanException e) {
			Throwable c = e.getCause();
			if (c instanceof IllegalArgumentException) {
				// most likely invalid worker name
				_error(c.getMessage(),null);
			} else {
				_error("Error during MBean invocation.",e.getCause());
			}
		} catch (Exception e) {
			_error("Failed to retrieve or invoke workers MBean.",e);
		}
	}

	private void usage() {
		_info("Usage:");
		_info("java -cp z.jar LaunchWorker <options> <workerName>");
		_info("Options:");
		_info("-gui\tOperate in graphical environments (w.r.t. messages)");
		_info("-? display this info");
		System.exit(1);
	}

	private void hello() {
		_info("Z2 Worker Launcher");
	}

	private void _info(String string) {
		System.err.println(string);
	}
	
	private void _error(String message, Throwable t) {
		if (gui) {
			JOptionPane.showMessageDialog(null, message+(t!=null? "\n["+t+"]":""),"Worker Launcher",JOptionPane.ERROR_MESSAGE);
		} else  {
			_info("SEVERE: "+message);
		}
		if (t!=null) {
			t.printStackTrace();
		}
	}

	public static void main(String[] args) throws Exception {
		new LaunchWorker().run(args);
	}

}
