/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.impl.gitcr.helper;

import java.io.File;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.jgit.api.Git;
import org.eclipse.jgit.lib.Repository;
import org.eclipse.jgit.storage.file.FileRepositoryBuilder;
import org.eclipse.jgit.transport.CredentialsProvider;
import org.eclipse.jgit.transport.TagOpt;
import org.eclipse.jgit.transport.URIish;

import com.zfabrik.components.provider.fs.AbstractFileSystemComponentRepository;
import com.zfabrik.impl.gitcr.GitComponentRepositoryImpl;
import com.zfabrik.util.fs.FileUtils;
import com.zfabrik.util.runtime.Foundation;

/**
 * The actual GitCR operations in complete for better testability
 */
public class GitCommand {
	private final static Logger logger = Logger.getLogger(GitComponentRepositoryImpl.class.getName());

	
	private File localRepo;
	private URIish originUri;
	private boolean isOptional;
	private String ref;
	private CredentialsProvider credentials;
	private int timeout;
	private String toString;
	
	public GitCommand(String toString, URIish originUri, File localRepo, String ref, boolean isOptional, CredentialsProvider credentials, int timeout) {
		super();
		this.toString = toString;
		this.originUri = originUri;
		this.localRepo = localRepo;
		this.ref = ref;
		this.isOptional = isOptional;
		this.credentials = credentials;
		this.timeout = timeout;
	}

	/*
	 * clone the repository and clean-up on failure
	 */
	public void doClone() {
		try {
			cloneRepository();
		} catch (Exception e) {
			// delete any repository ruins
			FileUtils.delete(this.localRepo);

			logger.log(Level.WARNING, "GIT-CR: Failed cloning " + this.toString(),e);

			switch (GitTools.isValidRepository(this.originUri)) {
			case invalid:
			case cannotTell:
				if (! GitTools.isOriginReachable(this.originUri)) {
					logger.info("GIT-CR: '" + this.originUri.getHost() + "' IS NOT reachable!");
				} else {
					logger.info("GIT-CR: '" + this.originUri.getHumanishName() + "' IS NOT a valid git repository on " + this.originUri.getHost() + "!");
				}
				break;

  			case valid:
  				// what could it be?
				logger.info("GIT-CR: '" + this.originUri.getHumanishName() + "' is valid git repository on " + this.originUri.getHost() + "!");
				break;
			}

			// something went wrong - maybe we are offline.
			if (this.isOptional) {
				logger.info("GIT-CR: '" + this.originUri.getHumanishName() + "' will be ignored because it's an optional repository or running in relaxed mode.");

			} else {
				throw new CloneFailedException(toString(),this.localRepo.getAbsolutePath(), e);
			}
		}
	}

	/*
	 * Clones the source-repository (defined by gitcr.uri) to the repository cache (defined by {@link AbstractFileSystemComponentRepository#getFileSystem()}).
	 * This method does not perform any checks whether a clone already exists etc.
	 */
	private void cloneRepository() throws Exception {
		Repository clonedRepo = null;

		logger.info("Cloning " + toString() + ". This will take some seconds...");

		// purging the folder allows to call clone in any situation
		FileUtils.delete(this.localRepo);

		try {
			// no clone available so far! Create it now...
			long tStart = System.currentTimeMillis();

			// clone the origin repository into the z2 work folder
			clonedRepo = GitTools.cloneRepository(this.originUri, this.localRepo, this.credentials, this.timeout);
			// check out the specified ref
			logger.fine("Checking out ref "+this.ref);
			GitTools.checkOutRef(clonedRepo,this.ref);

			logger.info("Created working-clone within " + (System.currentTimeMillis() - tStart) + " msec for " + toString());

		} finally {
			try { clonedRepo.close(); } catch (Exception e) { /* ignore */ }
		}
	}

	/**
	 * fetch the deltas and try clone on failure
	 */
	public void doFetch() {
		if (Foundation.isOfflineMode()) {
			logger.info("GIT-CR: '" + this.originUri.getHumanishName() + "' will be ignored (Running offline)");
			return;
		}
		Repository clonedRepo = null;
		try {
			clonedRepo = FileRepositoryBuilder.create(new File(this.localRepo, ".git"));
			fetchRepository(clonedRepo);
		} catch (Exception e) {

			// fetching failed!

			// just log a short description
			logger.log(Level.WARNING, "GIT-CR: Updating " + this.toString() + " failed with exception",e);

			switch (GitTools.isValidRepository(this.originUri)) {
			case invalid:
			case cannotTell:
				if (! GitTools.isOriginReachable(this.originUri)) {
					logger.info("GIT-CR: '" + this.originUri.getHost() + "' IS NOT reachable!");
				} else {
					logger.info("GIT-CR: '" + this.originUri.getHumanishName() + "' IS NOT a valid git repository on " + this.originUri.getHost() + "!");
				}

				if (this.isOptional) {
					// origin is not reachable, but repository is optional (maybe we are offline). Log this incident and try to continue
					logger.info("GIT-CR: '" + this.originUri.getHumanishName() + "' will be ignored because it's an optional repository or running in relaxed mode");

				} else {
					// throw exception in production mode
					throw new FetchFailedException(toString(), e);
				}

				break;

			case valid:
				// Origin is reachable which implies something went wrong locally. It could be for example that the fetch was successful but the merged failed.
				// In this case we will purge this clone and start from scratch.
				logger.info("GIT-CR: Will discard this clone and create a new one.");
				if (clonedRepo!=null) {
					clonedRepo.close();
				}
				doClone();
			}

		} finally {
			try { if (clonedRepo!=null) { clonedRepo.close();} } catch (Exception e) { /* ignore */ }
		}
	}

	/**
	 * fetches the deltas from the source-repository (defined by gitcr.uri) into the repository cache (defined by {@link AbstractFileSystemComponentRepository#getFileSystem()}).
	 * This method does not perform any checks whether a clone already exists etc.
	 */
	private void fetchRepository(Repository clonedRepo) throws Exception {
		// clone already exists. fetch the deltas
		logger.fine("Fetching deltas from " + toString());

		long tStart = System.currentTimeMillis();

		try (Git clonedGit = new Git(clonedRepo)) {

			// get the deltas from the origin repository
			clonedGit.
				fetch().
				setCredentialsProvider(this.credentials).	// set user/password if defined - can be null
				setTimeout(this.timeout).
				setTagOpt(TagOpt.FETCH_TAGS).
				call();
			
			// check out the specified ref (again)
			logger.fine("Checking out ref "+this.ref);
			GitTools.checkOutRef(clonedRepo,this.ref);

			logger.info("Fetched deltas within " + (System.currentTimeMillis() - tStart) + "msec from " + toString());
		}
	}

	@Override
	public String toString() {
		return "GIT-CR: "+(this.toString!=null? toString+", ":"")+"origin:" +this.originUri+ ",ref:" + this.ref + ",optional:"+this.isOptional;
	}

}
