/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.resources.util;

import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;
import java.lang.ref.SoftReference;
import java.lang.ref.WeakReference;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Queue;

/**
 * A map wrapper that manages a map over expirable values.
 *
 * Note: if an invalidation queue has been set, this map will NOT remove invalidated values from the
 * map but relies on external processing to do so.
 *
 * @author hb
 *
 * @param &lt;K&gt; key
 * @param &lt;V&gt; Value to hold (possibly by weak ref)
 * @param &lt;U&gt; Extra data associated with V but not determining its life cycle
 */
public class ExpirableValuesPseudoMap<K, V, U> {
	// management type per value
	public final static short HARD = 2;
	public final static short SOFT = 1;
	public final static short WEAK = 0;

	// queue
	private ReferenceQueue<V> queue = new ReferenceQueue<V>();
	private ValueHolder timed = null;

	// stats
	private int maxSize;
	private int intMaxSize;
	private int shrinks;

	public void resetStats() {
		this.shrinks = 0;
		this.maxSize = this.backingmap.size();
		this.intMaxSize = this.backingmap.size();
	}

	public int getMaxSize() {
		return maxSize;
	}

	public int getShrinks() {
		return shrinks;
	}

	public long getNextTime() {
		return nextTime;
	}



	// value holder (is the real in-memory key for managed resources
	//
	public class ValueHolder {

		// WeakValueHolder
		private class WeakValueHolderRef extends WeakReference<V> {
			public WeakValueHolderRef(V value) {
				super(value,queue);
			}
			public ValueHolder getValueHolder() {
				return ValueHolder.this;
			}
		}

		// SoftValueHolder
		private class SoftValueHolderRef extends SoftReference<V> {
			public SoftValueHolderRef(V value) {
				super(value,queue);
			}
			public ValueHolder getValueHolder() {
				return ValueHolder.this;
			}
		}

		// linking for timed list (a ring)
		private ValueHolder next,prev;

		private long ttl =0;
		private long untouch = Long.MAX_VALUE;
		private long expiration = Long.MAX_VALUE;
		private Reference<V> ref;
		private V value;
		private K key;
		private U extra;
		private int hashCode;

		public ValueHolder(K key, V value) {
			this(key, value,0,Long.MAX_VALUE,WEAK);
		}

		public ValueHolder(K key, V value,long ttl,long exp, short rt) {
			this.value = value;
			this.key = key;
			this.adjust(ttl, exp, rt);
			this.hashCode = value.hashCode();
		}

		public void adjust(long ttl,long exp, short rt) {
			this.ttl = ttl;
			this.expiration = exp;
			V v = this.getValue();
			switch (rt) {
			case WEAK:
				if (!(this.ref instanceof ExpirableValuesPseudoMap<?,?,?>.ValueHolder.WeakValueHolderRef)) {
					this.ref = new WeakValueHolderRef(v);
				}
				break;
			case SOFT:
				if (!(this.ref instanceof ExpirableValuesPseudoMap<?,?,?>.ValueHolder.SoftValueHolderRef)) {
					this.ref = new SoftValueHolderRef(v);
				}
				break;
			case HARD:
				if (this.ref!=null) {
					this.ref = null;
				}
				this.value = v;
				break;
			}
			this.touch();
			this.enqueue();
		}

		public V getValue() {
			if (this.value!=null)
				return this.value;
			if (this.ref!=null)
				return this.ref.get();
			return null;
		}

		public K getKey() {
			return this.key;
		}

		public U getExtra() {
			return extra;
		}

		public void setExtra(U extra) {
			this.extra = extra;
		}

		public void touch() {
			if (this.ttl>0) {
				this.untouch = System.currentTimeMillis()+this.ttl;
				if (this.value==null) {
					this.value = getValue();
				}
			} else {
				this.untouch = Long.MAX_VALUE;
				if (this.ref!=null) {
					this.value=null;
				}
			}
		}

		private long nextEvent() {
			return (this.untouch<this.expiration? this.untouch:this.expiration);
		}

		private void enqueue() {
			if ((this.expiration<Long.MAX_VALUE) || (this.untouch<Long.MAX_VALUE)) {
				// we have timely behavior
				if (this.prev==null) {
					// put in queue first
					if (ExpirableValuesPseudoMap.this.timed==null) {
						// first entry
						ExpirableValuesPseudoMap.this.timed = this;
						this.prev = this;
						this.next = this;
					} else {
						// put at beginning
						this.prev = ExpirableValuesPseudoMap.this.timed;
						this.next = this.prev.next;
						this.next.prev = this;
						this.prev.next = this;
						ExpirableValuesPseudoMap.this.timed = this;
					}
				}
				// find best place
				long l = this.nextEvent();
				// first down:
				ValueHolder c = this.next;
				while ((c!=this) && (c.nextEvent()<l)) c=c.next;
				c = c.prev;
				if (c != this) {
					// since c.nextEvent>=l, we move before c.
					c.prev.next = this;
					this.prev = c.prev;
					c.prev = this;
					this.next = c;
					if (ExpirableValuesPseudoMap.this.timed==c) {
						ExpirableValuesPseudoMap.this.timed = this;
					}
				} else {
					// didn't move. Check the other direction
					c = this.prev;
					while ((c!=this) && (c.nextEvent()>l)) c=c.prev;
					c = c.next;
					if (c != this) {
						// now c.nextEvent <= l, so we move behind c
						c.next.prev = this;
						this.next = c.next;
						c.next = this;
						this.prev = c;
					}
				}
			} else {
				dequeue();
			}
		}

		private void dequeue() {
			if (this.next==null) return;
			if (this.next == this) {
				// only one...
				ExpirableValuesPseudoMap.this.timed = null;
			} else {
				if (ExpirableValuesPseudoMap.this.timed==this) {
					ExpirableValuesPseudoMap.this.timed = this.next;
				}
				// remove from queue
				this.prev.next = this.next;
				this.next.prev = this.prev;
			}
			this.prev = null;
			this.next = null;
		}

		public int hashCode() {
			return this.hashCode;
		}

		public boolean equals(Object obj) {
			// reduced to value equality
			if (obj==this) return true;
			if (!(obj instanceof ExpirableValuesPseudoMap<?,?,?>.ValueHolder)) return false;
			@SuppressWarnings("unchecked")
			V v1 = this.getValue(), v2 = ((ValueHolder)obj).getValue();
			if (v1==null) {
				return v2==null;
			}
			return v1.equals(v2);
		}

		private void tick(long now) {
			if (now>=this.untouch) {
				this.value = null;
			}
			if (now>=this.expiration) {
				// list in expiration queue
				if (ExpirableValuesPseudoMap.this.expirationQueue!=null) {
					ExpirableValuesPseudoMap.this.expirationQueue.add(this);
				}
			}
		}

		public long getTtl() {
			return ttl;
		}

		public long getExpiration() {
			return expiration;
		}

		public short getRefMode() {
			if (this.ref==null) {
				return HARD;
			}
			if (this.ref instanceof ExpirableValuesPseudoMap<?,?,?>.ValueHolder.WeakValueHolderRef) {
				return WEAK;
			}
			if (this.ref instanceof ExpirableValuesPseudoMap<?,?,?>.ValueHolder.SoftValueHolderRef) {
				return SOFT;
			}
			throw new IllegalStateException("invalid ref assignment");
		}
	}

	/// END ValueHolder

	private HashMap<K, ValueHolder> backingmap;
	private Map<K, ValueHolder> immutableBackingMapView;
	private Queue<ValueHolder> expirationQueue, collectedQueue;

	public ExpirableValuesPseudoMap() {
		this.backingmap = new HashMap<K, ValueHolder>();
		this.immutableBackingMapView = Collections.unmodifiableMap(this.backingmap);
	}

	public ExpirableValuesPseudoMap(int initialSize) {
		this.backingmap = new HashMap<K, ValueHolder>(initialSize);
		this.immutableBackingMapView = Collections.unmodifiableMap(this.backingmap);
	}

	public Queue<ValueHolder> setExpirationQueue(Queue<ValueHolder> q) {
		Queue<ValueHolder> r = this.expirationQueue;
		this.expirationQueue = q;
		return r;
	}

	public Queue<ValueHolder> setInvalidationQueue(Queue<ValueHolder> q) {
		Queue<ValueHolder> r = this.collectedQueue;
		this.collectedQueue = q;
		return r;
	}

	public Map<K,ValueHolder> map() {
		return this.immutableBackingMapView;
	}


	//
	// our special modifying methods
	//
	public ValueHolder  put(K key, V value) {
		if (key==null) {
			throw new NullPointerException("Resource key cannot be null");
		}
		this.countDown--;
		tick();
		ValueHolder vh = new ValueHolder(key,value);
		this.backingmap.put(key, vh);
		int l = this.backingmap.size();
		if (l>this.maxSize)
			this.maxSize = l;
		if (l>this.intMaxSize)
			this.intMaxSize = l; // for shrinks
		return vh;
	}

	public V remove(K key) {
		this.countDown--;
		tick();
		ValueHolder vh = this.backingmap.remove(key);
		if (vh!=null) {
			vh.dequeue();
			return vh.getValue();
		}
		return null;
	}

	public V get(K key) {
		ValueHolder vh = this.backingmap.get(key);
		if (vh!=null)
			return vh.getValue();
		return null;
	}

	public void setTickTimeDelta(int n) {
		this.delta = n;
	}

	public void setTickOpCount(int n){
		this.opcount = n;
	}
	//
	// upon a tick we do expirations and cleanouts. This map implementation needs external ticks to do its work
	//

	private int delta = 500;
	private int opcount = 500;
	private long nextTime = System.currentTimeMillis()+delta;
	private int  countDown = opcount;

	@SuppressWarnings("unchecked")
	public void tick() {
		// process ref queue
		Reference<? extends V> r;
		ValueHolder vh;
		while ((r=this.queue.poll())!=null) {
			if (r instanceof ExpirableValuesPseudoMap<?,?,?>.ValueHolder.WeakValueHolderRef) {
				vh = ((ValueHolder.WeakValueHolderRef) r).getValueHolder();
			} else
			if (r instanceof ExpirableValuesPseudoMap<?,?,?>.ValueHolder.SoftValueHolderRef) {
				vh = ((ValueHolder.SoftValueHolderRef) r).getValueHolder();
			} else
				throw new IllegalStateException("internal error: unknown ref type in queue");
			vh.dequeue();
			if (this.collectedQueue!=null) {
				// remove from map must be done in external ticker!
				this.collectedQueue.add(vh);
			} else {
				this.remove(vh.getKey());
			}
		}
		if ((countDown<=0) || (nextTime<=System.currentTimeMillis())) {
			if (this.timed!=null) {
				// process timed holders
				// collect first.
				long now = System.currentTimeMillis();
				int n = 0;
				ValueHolder c = this.timed;
				do {
					if (c.nextEvent()<=now) n++;
					c = c.next;
				} while (c!=this.timed);
				Object[] l = new Object[n];
				c = this.timed;
				for (int i=0;i<n;i++) {
					l[i]=c;
					c=c.next;
				}
				// now process
				for (int i=0;i<n;i++) {
					((ValueHolder) l[i]).tick(now);
				}
			}
			countDown = opcount;
			nextTime  = System.currentTimeMillis()+delta;

			int l = this.backingmap.size();
			int i=1; while ((1<<i)<l) i++;
			int j=1; while ((1<<j)<intMaxSize) j++;
			if ((j-i)>=2) {
				// shrink
				HashMap<K, ValueHolder> nm = new HashMap<K, ValueHolder>(l);
				nm.putAll(this.backingmap);
				// redefine maps
				this.backingmap = nm;
				this.immutableBackingMapView = Collections.unmodifiableMap(this.backingmap);
				this.shrinks++;
				this.intMaxSize=l;
			}
		}
	}

}
