/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.util.threading;

import java.security.AccessController;
import java.security.PrivilegedAction;
import java.security.PrivilegedExceptionAction;
import java.util.concurrent.Callable;

/**
 * Threading utils. Not thread pools and such but rather utilities that help
 * isolating resources and avoiding memory leaks.
 * 
 * @author hb
 *
 */
public class ThreadUtil {

	/**
	 * Stupid... but there is places all over that hold on to
	 * AccessControlContexts and Context Class Loaders. This easily consitutes a
	 * leak. So, for stuff that doesn't need anything from the context class
	 * loader this is a way of keeping context to a minimum;
	 * 
	 * Execute a callable with a clean thread context and security context as
	 * to avoid any passing on of thread context and security context to another
	 * thread that may be spawned from here and may end up holding copies in the end.
	 * Any exception thrown will be propagated.
	 * 
	 * @param contextLoader A class loader to set as context class loader 
	 * @param callable A {@link Callable} to invoke
	 * @return Return value from the {@link Callable} invocation
	 * @throws Exception
	 */
	public static <T> T cleanContextExceptionExecute(ClassLoader contextLoader, final Callable<T> callable) throws Exception {
		ClassLoader cl = Thread.currentThread().getContextClassLoader();
		Thread.currentThread().setContextClassLoader(contextLoader);
		try {
			return AccessController.doPrivileged(new PrivilegedExceptionAction<T>() {
				public T run() throws Exception {
					return callable.call();
				}
			});
		} finally {
			Thread.currentThread().setContextClassLoader(cl);
		}
	}

	/**
	 * Short version of {@link #cleanContextExceptionExecute(ClassLoader, Callable)} passing in <code>null</code>
	 * as context class loader.
	 * 
	 * @param callable A {@link Callable} to invoke
	 * @return Return value from the {@link Callable} invocation
	 * @throws Exception
	 */
	public static <T> T cleanContextExceptionExecute(final Callable<T> callable) throws Exception {
		return cleanContextExceptionExecute(null, callable);
	}

	/**
	 * Execute a callable with a clean thread context and security context as
	 * to avoid any passing on of thread context and security context to another
	 * thread that may be spawned from here and may end up holding copies in the end.
	 * Any exception thrown will be forwarded via a generic runtime exception.
	 * 
	 * @param contextLoader A class loader to set as context class loader 
	 * @param callable A {@link Callable} to invoke
	 * @return Return value from the {@link Callable} invocation
	 */
	public static <T> T cleanContextExecute(ClassLoader contextLoader, final Callable<T> callable) {
		ClassLoader cl = Thread.currentThread().getContextClassLoader();
		Thread.currentThread().setContextClassLoader(contextLoader);
		try {
			return AccessController.doPrivileged(new PrivilegedAction<T>() {
				public T run() {
					try {
						return callable.call();
					} catch (Exception e) {
						throw new RuntimeException(e);
					}
				}
			});
		} finally {
			Thread.currentThread().setContextClassLoader(cl);
		}
	}
	
	/**
	 * Short version of {@link #cleanContextExecute(ClassLoader, Callable)} passing in <code>null</code>
	 * as context class loader.
	 * 
	 * @param callable A {@link Callable} to invoke
	 * @return Return value from the {@link Callable} invocation
	 */
	public static <T> T cleanContextExecute(final Callable<T> callable) {
		return cleanContextExecute(null, callable);
	}

}
