/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.workers.home;

import java.io.IOException;
import java.io.Serializable;
import java.util.Collection;
import java.util.Map;

import com.zfabrik.launch.ProcessRunner;
import com.zfabrik.util.runtime.Foundation;
import com.zfabrik.workers.worker.HomeHandle;

/**
 * Interface supported by a component of type <code>com.zfabrik.homeLayout</code>. 
 * <p>
 * When starting a z2 environment in server mode (as compared to embedded, see {@link ProcessRunner})
 * the home layout determines the worker processes (see {@link IWorkerProcess}) to be started and
 * managed by the home process.
 * <p>
 * A home layout component has no further resources. Typically a home layout definition looks 
 * like the following properties set:
 * <pre>
 * com.zfabrik.component.type=com.zfabrik.homeLayout
 * home.workers=&lt;name of worker process component 1&gt;, &lt;name of worker process component 2&gt;, ... 
 * </pre>
 * During startup, the home process identifies its home layout via the system property
 * <code>com.zfabrik.home.layout</code> (see also {@link Foundation#HOME_LAYOUT_COMPONENT}).
 * The home layout (see above) provides the list of worker processes to manage. See 
 * {@link IWorkerProcess} for more information about worker processes.
 * <p>
 * This interface allows to send messages to worker processes from the home process. It is
 * hence mostly used internally and should not be used be applications.
 * <p>
 * See {@link HomeHandle} for a worker process perspective onto the running
 * home layout.
 *  
 * 
 * @author hb
 *
 */
public interface IHomeLayout {
	String TYPE = "com.zfabrik.homeLayout";
	
	/**
	 * broadcast scope: local home process
	 */
	int SCOPE_HOME = 1;
	/**
	 * broadcast scope: all worker processes in home
	 */
	int SCOPE_WORKERS = 2;
	/**
	 * broadcast scope: exclude the sending worker process
	 */
	int SCOPE_OTHERS  = 4;
	/**
	 * broadcast scope: send invalidation out to the cluster
	 */
	int SCOPE_CLUSTER = 8;
	/**
	 * broadcast scope: local home process
	 */
	int SCOPE_GLOBAL = SCOPE_HOME | SCOPE_WORKERS | SCOPE_CLUSTER;


	/**
	 * broadcast a message to all worker processes of a home. Timeouts will result in exceptions 
	 * @param args
	 * @param timeout timeout per worker or -1 for default
	 * @return
	 * @throws IOException
	 */
	void broadcastMessage(Map<String, Serializable> args, long timeout) throws IOException;

	/**
	 * broadcast a message to all worker processes of a home 
	 * @param args
	 * @param timeout timeout per worker or -1 for default
	 * @param killOnTimeOut if <code>true</code> a time out will lead to a worker process termination.
	 * @return
	 * @throws IOException
	 */
	void broadcastMessage(Map<String, Serializable> args, long timeout,boolean killOnTimeOut) throws IOException;

	/**
	 * broadcast resource invalidations to all worker processes of a home and, if specified, to all 
	 * homes and workers in a cluster. Invalidations will be processed on the home process as well and before
	 * distribution to worker processes.
	 * 
	 * @param args
	 * @param timeout timeout per worker or -1 for default
	 * @return
	 * @throws IOException
	 */
	void broadcastInvalidations(Collection<String> invs, long timeout, int scope, String senderWorker) throws IOException;

	void start();
}
