/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.impl.cluster;

import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.sql.DataSource;

import com.zfabrik.cluster.IInvalidator;
import com.zfabrik.components.IComponentDescriptor;
import com.zfabrik.components.IComponentsLookup;
import com.zfabrik.components.IComponentsManager;
import com.zfabrik.components.IDependencyComponent;
import com.zfabrik.impl.cluster.dao.InvalidationsDAO;
import com.zfabrik.resources.IResourceHandle;
import com.zfabrik.resources.ResourceBusyException;
import com.zfabrik.resources.provider.Resource;
import com.zfabrik.util.runtime.Foundation;
import com.zfabrik.util.threading.TimerUtil;
import com.zfabrik.work.ApplicationThreadPool;
import com.zfabrik.workers.home.IHomeLayout;
import com.zfabrik.workers.home.IWorkersManager;

public class InvalidatorResource extends Resource implements IDependencyComponent, IInvalidator {
	private static final int MAXPULL = 100;

	private final static String DATASOURCE = "invalidator.dataSourceComponent";
	private final static String TBL_PREFIX = "invalidator.tablePrefix";
	private final static String CHECKDELAY = "invalidator.delay";

	private String name;
	private InvalidationsDAO dao;
	private long delay;
	private CheckerTT task;
	private Timer timer;

	public InvalidatorResource(String name) {
		this.name = name;
	}

	public synchronized <T> T as(Class<T> clz) {
		if (IDependencyComponent.class.equals(clz)) {
			return clz.cast(this);
		} else if (IInvalidator.class.equals(clz)) {
			_load();
			return clz.cast(this);
		}
		return null;
	}

	public void init() {
	}

	public synchronized void invalidate() throws ResourceBusyException {
		_unload();
	}

	public synchronized void prepare() {
		_load();
	}

	private void _load() {
		if (this.dao == null) {
			IComponentDescriptor desc = IComponentsManager.INSTANCE.getComponent(this.name);
			if (desc == null) {
				throw new IllegalStateException("Invalidator component not found: " + this.name);
			}
			String dels = desc.getProperty(CHECKDELAY);

			if (dels == null) {
				throw new IllegalStateException("Invalidator component must specify delay between checks (" + CHECKDELAY + "): " + this.name);
			}
			try {
				this.delay = Long.parseLong(dels.trim());
			} catch (NumberFormatException nfe) {
				throw new IllegalStateException("Illegal specifcation of delay between checks (" + dels + "): " + this.name, nfe);
			}

			String dsn = desc.getProperty(DATASOURCE);
			if (dsn == null) {
				throw new IllegalStateException("Invalidator component must specify data source component (" + DATASOURCE + "): " + this.name);
			}
			IResourceHandle rh = IComponentsLookup.INSTANCE.lookup(dsn, IResourceHandle.class);
			DataSource ds = rh.as(DataSource.class);
			if (ds == null) {
				throw new IllegalStateException("Data Source component (" + dsn + ") does not provide " + DataSource.class.getName() + " instance: "
						+ this.name);
			}
			handle().addDependency(rh);
			try {
				this.dao = new InvalidationsDAO(ds, desc.getProperty(TBL_PREFIX));
			} catch (Exception e) {
				throw new IllegalStateException("Failed to initialize cluster invalidator: " + this.name, e);
			}
			if (!Foundation.isWorker()) {
				// on home, we run recurring invalidation checks
				this.task = new CheckerTT();
				this.timer = TimerUtil.createTimer("Invalidations polling ("+name+")",true);
				this.timer.schedule(this.task, this.delay, this.delay);
			}
			handle().adjust(0, Long.MAX_VALUE, IResourceHandle.HARD);
		}
	}

	private void _unload() {
		if (this.dao != null) {
			try {
				if (this.timer != null) {
					this.task.invalidate();
					this.timer.cancel();
				}
			} finally {
				this.dao = null;
				this.timer = null;
				this.task = null;
				handle().adjust(0, Long.MAX_VALUE, IResourceHandle.WEAK);
			}
		}
	}

	public void postInvalidations(Collection<String> invs) {
		if (this.dao != null) {
			try {
				if (logger.isLoggable(Level.FINE)) {
					logger.fine("posting " + invs.size() + " invalidations: " + invs);
				}
				this.dao.post(invs);
			} catch (Exception e) {
				throw new RuntimeException("Failed to post invalidations", e);
			}
		}
	}

	// ------------ timer task -----------------//
	private class CheckerTT extends TimerTask {
		private final Checker checker = new Checker();

		public void invalidate() {
			this.checker.invalidate();
		}

		public void run() {
			synchronized (this) {
				ApplicationThreadPool.instance().execute(true, checker);
			}
		}

	}

	// ------------ application runnable task -----------------//
	private class Checker implements Runnable {
		private boolean invalid;

		public synchronized void invalidate() {
			this.invalid = true;
		}

		public synchronized boolean isInvalid() {
			return this.invalid;
		}

		@Override
		public void run() {
			// do our thing here...
			// 1. read a number of invalidations
			// 2. execute them
			// 3. if we got invalidated, or if it was the last ones, run a
			// verify
			if (logger.isLoggable(Level.FINER))
				logger.finer("Starting check for invalidations: " + name);
			boolean left = true;
			while (left && (!isInvalid())) {
				List<String> invs = Collections.emptyList();
				try {
					invs = dao.pull(MAXPULL);
				} catch (Exception e1) {
					logger.log(Level.WARNING, "Failed to pull distributed invalidations: " + name, e1);
				}
				if (invs.size() > 0) {
					if (logger.isLoggable(Level.FINE))
						logger.fine("Found " + invs.size() + " invalidations: ("+invs+"):" + name);
					left = invs.size() == MAXPULL;
					// run invalidations broadcast to workers.
					try {
						IWorkersManager.INSTANCE.getActiveLayout().broadcastInvalidations(invs, -1, IHomeLayout.SCOPE_HOME | IHomeLayout.SCOPE_WORKERS,null);
					} catch (IOException e) {
						// TODO should we actually leave invalidations on hold,
						// if we cannot broadcast?
						logger.log(Level.WARNING, "Failed to broadcast distributed invalidations: " + name, e);
					}
				} else {
					left = false;
				}
			}
		}
	}

	private final static Logger logger = Logger.getLogger(InvalidatorResource.class.getName());
}
