/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.impl.cluster.dao;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.zfabrik.util.runtime.Foundation;

/**
 * Never forget: create your databased with UTF8 default encoding. (create database table1 default character set 'utf8', default collate 'utf8_general_ci')
 * @author hb
 *
 */
public class InvalidationsDAOForMySQL implements InvalidationsImpl {
	private static final int MYSQL_UNKNOWN_TABLE = 1051;

	private final static int IMPLEMENTATION_VERSION = 4;
	private static final String REVISION = InvalidationsDAOForMySQL.class.getName()+"/revision";
	
	private String tinvs, ttops, tprops;

	public void init(Connection c, String tablePrefix) throws Exception {
		tinvs = tablePrefix+"INVALIDATIONS";
		ttops = tablePrefix+"INVTOP";
		tprops= tablePrefix+"PROPERTIES";

		// try to read the customizing
		Properties p = new Properties();
		boolean updateSchema=true;
		try {
			_readProps(c,p);
			String vs = p.getProperty(REVISION);
			if (vs!=null) {
				try {
					if (IMPLEMENTATION_VERSION==Integer.parseInt(vs)) {
						updateSchema = false;
					}
				} catch (NumberFormatException nfe) {}
			}
		} catch (SQLException sqle) {	}
		
		if (updateSchema) {
			createSchema(c);
			p.setProperty(REVISION,Integer.toString(IMPLEMENTATION_VERSION));
			_writeProps(c,p);
		}

		// check tops
		long max =0;
		PreparedStatement ps = c.prepareStatement("select max(SEQ) from "+tinvs);
		ResultSet rs = ps.executeQuery();
		if (rs.next()) {
			max = rs.getLong(1);
		}
		rs.close();
		ps.close();
		
		ps = c.prepareStatement("insert into "+ttops+" (TOP,HOME) values (?,?) on duplicate key update TOP=?");
		ps.setLong(1,max);
		ps.setString(2,_getHomeCluster());
		ps.setLong(3,max);
		ps.executeUpdate();
		ps.close();
	}

	private void _writeProps(Connection c, Properties p) throws SQLException {
		PreparedStatement ps = c.prepareStatement("insert into "+tprops+" (NAME, VALUE) values (?, ?) 	on duplicate key update VALUE=?");
		try {
			for (Map.Entry<Object,Object> e:p.entrySet()) {
				ps.setString(1, (String) e.getKey()); 
				ps.setString(2, (String) e.getValue()); 
				ps.setString(3, (String) e.getValue());
				ps.execute();
			}
		} finally {
			ps.close();
		}
	}

	private void _readProps(Connection c,Properties p) throws SQLException {
		Statement s = c.createStatement();
		try {
			ResultSet rs = s.executeQuery("select NAME,VALUE from "+tprops);
			try {
				while (rs.next()) {
					p.setProperty(rs.getString(1),rs.getString(2));
				}
			} finally {
				rs.close();
			}
		} finally {
			s.close();
		}
	}

	private void createSchema(Connection c) throws Exception {
		Statement s = c.createStatement();
		try {
			_drop(s,tinvs);
			_drop(s,ttops);
			_drop(s,tprops);
	
			_create(s, tinvs,	
								"SEQ bigint(20) NOT NULL auto_increment,"+
								"RESOURCE char(255) NOT NULL,"+
								"ISSUER char(200) NOT NULL,"+
								"TIME timestamp NOT NULL default CURRENT_TIMESTAMP on update CURRENT_TIMESTAMP,"+
								"PRIMARY KEY  (`SEQ`)"
			);
			_create(s, ttops,	"HOME char(200) NOT NULL,TOP bigint(20) NOT NULL, PRIMARY KEY  (`HOME`)");
			
			_create(s, tprops, "NAME char(200) NOT NULL,VALUE char(255) default NULL, PRIMARY KEY  (`NAME`)");
		} finally {
			s.close();
		}
	}

	private void _create(Statement s, String table, String def) throws Exception {
		s.executeUpdate("create table "+table+" ("+def+") ENGINE=InnoDB DEFAULT CHARSET=utf8");
	}

	private void _drop(Statement s, String table) throws Exception {
		try {
			s.executeUpdate("drop table "+table);
		} catch (SQLException e) {
			if (e.getErrorCode()!=MYSQL_UNKNOWN_TABLE) {
				logger.log(Level.WARNING, "unexpected error when dropping table",e);
			}
		}
	}

	
	public Properties getCustomizing(Connection c) throws Exception {
		Properties p = new Properties();
		_readProps(c, p);
		return p;
	}
	
	@Override
	public void updateCustomizing(Connection c, Properties p) throws Exception {
		_writeProps(c, p);
	}


	public void post(Connection c, List<String> invs) throws Exception {
		
		PreparedStatement ps = c.prepareStatement("insert into "+tinvs+" (RESOURCE, ISSUER) values (?,?)");
		
		String homeId = _getHomeCluster();
		for (String s:invs) {
			ps.setString(1,s);
			ps.setString(2,homeId);
			ps.executeUpdate();
		}
		
	}

	public List<String> pull(Connection c, int max) throws Exception {
		// we use the time stamp to make sure we do not even care for anything older than
		// our start time with a configured uncertainty.
		// we use our TOP to limit the sequence from below further.
		// after we have pulled, we delete all invalidations older than 10min.
	
		List<String> res = new ArrayList<String>(max);
		
		String homeId = _getHomeCluster();
		
		PreparedStatement ps = c.prepareStatement(
				"select RESOURCE,SEQ from "+tinvs+" where  SEQ>(select TOP from "+ttops+" where HOME=?) and ISSUER<>? order by SEQ asc"
		);
		ps.setString(1, homeId);
		ps.setString(2, homeId);
		ps.setFetchSize(max);
		if (logger.isLoggable(Level.FINER))
			logger.finer("executing "+ps);
		ResultSet rs = ps.executeQuery();
		long maxSeq = -1,h;
		try {
			while (rs.next()) {
				 res.add(rs.getString(1));
				 h = rs.getLong(2);
				 if (h>maxSeq) maxSeq=h;
			}
		} finally {
			rs.close();
		}

		if (maxSeq>-1) {
			if (logger.isLoggable(Level.FINE)) {
				logger.fine("New top for home "+homeId+" is "+maxSeq);
			}
			ps = c.prepareStatement("update "+ttops+"  set TOP=? where HOME=?");
			ps.setString(2,homeId);
			ps.setLong(1,maxSeq);
	
			if (logger.isLoggable(Level.FINER))
				logger.finer("executing "+ps);
			if (ps.executeUpdate()==0) {
				if (logger.isLoggable(Level.FINE))
					logger.fine("no top for home "+homeId+" entry in table. Inserting new entry");
				ps = c.prepareStatement("insert into "+ttops+" (HOME, TOP) values (?, ?)");
				ps.setString(1,homeId);
				ps.setLong(2,maxSeq);
				if (logger.isLoggable(Level.FINER))
					logger.finer("executing "+ps);
				ps.executeUpdate();
			}
		}
		// delete old entries
		ps = c.prepareStatement("delete from "+tinvs+" where TIME <  timestampadd(minute,-5,now())");
		if (logger.isLoggable(Level.FINER))
			logger.finer("executing "+ps);
		ps.executeUpdate();
		return res;
	}

	private String _getHomeCluster() {
		String homeId = Foundation.getProperties().getProperty(Foundation.HOME_CLUSTER);
		return homeId;
	}

	private final static Logger logger = Logger.getLogger(InvalidationsDAOForMySQL.class.getName());
}
