/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.impl.workers.home;

import java.io.IOException;
import java.io.Serializable;
import java.lang.management.ManagementFactory;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.StandardMBean;

import com.zfabrik.cluster.IInvalidator;
import com.zfabrik.components.IComponentDescriptor;
import com.zfabrik.components.IComponentsLookup;
import com.zfabrik.components.IComponentsManager;
import com.zfabrik.impl.workers.MessageDistributionException;
import com.zfabrik.impl.workers.WorkerConstants;
import com.zfabrik.impl.workers.home.io.WorkerNotifier;
import com.zfabrik.resources.IResourceHandle;
import com.zfabrik.resources.IResourceManager;
import com.zfabrik.resources.ResourceBusyException;
import com.zfabrik.resources.provider.Resource;
import com.zfabrik.sync.SynchronizationRunner;
import com.zfabrik.util.expression.X;
import com.zfabrik.work.ApplicationThreadPool;
import com.zfabrik.work.WorkUnit;
import com.zfabrik.workers.IMessageHandler;
import com.zfabrik.workers.home.IHomeLayout;
import com.zfabrik.workers.home.IWorkerProcess;

/**
 * The home layout knows all currently running workers (since it is the one starting them). It knows how to broadcast 
 * messages and how to re-establish a basic home layout (i.e. without those dynamically started workers)
 */
public class HomeLayoutResource extends Resource implements IHomeLayout, HomeLayoutMBean {
	private static final X X_DEFAULT_INVALIDATOR = X.var(IComponentDescriptor.COMPONENT_TYPE).eq(X.val(IInvalidator.TYPE)).and(
								X.var(IInvalidator.INVALIDATOR_ID).eq(X.val(IInvalidator.INVALIDATOR_ID_DEFAULT)));

	public volatile static IResourceHandle INSTANCE;

	private String name;
	private final static String WORKERS = "home.workers";
	private ObjectName on;
	private String invalidator;
	private long invalidationCount;
	
	
	public HomeLayoutResource(String name) {
		this.name = name;
	}

	public void init() {}
	
	public <T> T as(Class<T> clz) {
		if (IHomeLayout.class.equals(clz)) {
			if (INSTANCE != null &&  INSTANCE!=handle()) {
				throw new IllegalStateException("There can only be one home layout used at a time (currently " + INSTANCE.getResourceInfo().getName() + "): " + name);
			} else {
				INSTANCE = handle();
			}
			return clz.cast(this);
		}
		return null;
	}

	@Override
	public void invalidate() throws ResourceBusyException {
		this.invalidationCount++;
		INSTANCE=null;
		stop();
	}
	
	private synchronized long getInvalidationCount() {
		return this.invalidationCount;
	}

	// clean out old workers and return a list of known workers.
//	private Set<IResourceHandle> _getAllWorkerProcesses() {
//		Set<IResourceHandle> res = new HashSet<IResourceHandle>(this.workerProcesses.size());
//		res.addAll(this.workerProcesses);
//		return res;		
//	}

	public synchronized void startWorker(final String componentName) throws Exception {
		ApplicationThreadPool.instance().executeAs(new Runnable() {
			public void run() {
				IResourceHandle rh = IComponentsLookup.INSTANCE.lookup(componentName, IResourceHandle.class);;
				IWorkerProcess wp = rh.as(IWorkerProcess.class);
				if (wp==null)
					throw new IllegalArgumentException("Worker component ("+componentName+") not found!");
				wp.start();
				rh.addDependency(handle());
			}
		},true);
	}
	
	/**
	 * TODO: concurrent stop
	 */
	private void stop() {
		synchronized (this) {
			if (this.on!=null) {
				try {
					ManagementFactory.getPlatformMBeanServer().unregisterMBean(this.on);
				} catch (Exception e) {
					logger.log(Level.WARNING,"Problem unregistering home layout MBean: "+this.name);
				} finally {
					this.on=null;
				}
			}
		}
	}
	/**
	 * TODO: concurrent start TODO: workers should indicate what layout they
	 * belong to and not the other way around
	 */
	public void start() {
		synchronized (this) {
			IComponentDescriptor desc = IComponentsManager.INSTANCE.getComponent(this.name);
			Collection<String> res;
			try {
				res = IComponentsManager.INSTANCE.findComponents(X_DEFAULT_INVALIDATOR);
			} catch (IOException e) {
				throw new RuntimeException("Failed to search for default invalidator: "+this.name,e);
			} 
							
			if (res.size()>1) 
				throw new IllegalStateException("Found more than one default invalidator ("+res+"): "+this.name);
			if (res.size()==1) {
				this.invalidator = res.iterator().next();
			}
			if (this.invalidator!=null) {
				// load the invalidator
				IInvalidator in = IComponentsLookup.INSTANCE.lookup(this.invalidator, IInvalidator.class);
				if (in==null) {
					throw new IllegalStateException("Failed to load invalidator implementation ("+this.invalidator+"): "+name);
				}
			}
			String workers = desc.getProperty(WORKERS);
			if (workers != null) {
				StringTokenizer tok = new StringTokenizer(workers, ",");
				int l = tok.countTokens(),i=0;
				Runnable[] starters = new Runnable[l];
				while (tok.hasMoreElements()) {
					final String worker = tok.nextToken().trim();
					starters[i++] = new Runnable() {
						public void run() {
							IResourceHandle rh = IComponentsLookup.INSTANCE.lookup(worker, IResourceHandle.class);
							IWorkerProcess  wp = rh.as(IWorkerProcess.class);
							if (wp!=null) {
								// start (unless already running)
								wp.start();
								// worker depends on the layout
								rh.addDependency(handle());
							} else {
								logger.warning("Worker process ("+worker+") not found");
							}
						}
					};
				}
				// uncomment here for concurrent start
//				ApplicationThreadPool.instance().execute(true, starters);
				// uncomment here for sequential start. Note: The worker process will not reply to the initial ping until the 
				// worker target state have been reach and its runnable (if specified) been executed.
				for (Runnable r : starters) {
					r.run();
				}
			}
		}
		if (this.on==null) {
			try {
				this.on = ObjectName.getInstance("zfabrik:type="+HomeLayoutResource.class.getName());
				MBeanServer mbs = ManagementFactory.getPlatformMBeanServer();
				if (mbs.isRegistered(this.on)) {
					mbs.unregisterMBean(this.on);
				} 
				mbs.registerMBean(new StandardMBean(this, HomeLayoutMBean.class),this.on);
			} catch (Exception e) {
				throw new IllegalStateException("Failed to register MBean for home layout: "+this.name,e);
			}
		}
	}

	public void broadcastInvalidations(Collection<String> invs, long timeout, int scope, String senderWorker) throws IOException {
		if (logger.isLoggable(Level.FINE)) {
			logger.fine("processing "+invs.size()+" invalidations. Scope is "+(scope==SCOPE_CLUSTER?"cluster":"home"));
		}
		// 1. post to cluster
		if ((scope & SCOPE_CLUSTER)!=0 && this.invalidator!=null) {
			IInvalidator in = IComponentsLookup.INSTANCE.lookup(this.invalidator, IInvalidator.class);
			if (in!=null) {
				in.postInvalidations(invs);
			} 
			// commit the current workunit, as to make sure invalidations are spread
			try {
				WorkUnit.commitCurrent();
			} catch (Exception e) {
				throw new RuntimeException("Error during commit of invalidations: "+name,e);
			}
		}
		long c = getInvalidationCount();
		if ((scope & ~SCOPE_CLUSTER)!=0) {
			SynchronizationRunner r=null;
			try {
				if ((scope & SCOPE_HOME)!=0) {
					// acquire sync runner here... down there our class loader might already reject giving it to us.
					r = new SynchronizationRunner(SynchronizationRunner.VERIFY_ONLY);
					// 2. process local invalidations
					IResourceManager.INSTANCE.invalidate(invs);
				}		
				// 3. unless we have been invalidated (which implies that all workers have been stopped), we broadcast to them
				if ((scope & SCOPE_WORKERS)!=0 && c==getInvalidationCount()) {
					Map<String,Serializable> args = new HashMap<String, Serializable>();
					args.put(IMessageHandler.COMMAND, WorkerConstants.COMMAND_INVALIDATE);
					args.put(WorkerConstants.INVALIDATIONSET, (Serializable)invs);
					try {
						_broadcastMessage(args,timeout, ((scope & SCOPE_OTHERS)!=0? senderWorker:null),false);
					} catch (IOException e) {
						throw new IllegalStateException("Error during send of invalidation message to worker processes",e);
					}
				}
			} finally {
				// schedule  a verification
				if ((scope & SCOPE_HOME)!=0) {
					r.execute(false);
				}
			}
		}
	}

	private void _broadcastMessage(Map<String, Serializable> args, long timeout, String exclude,boolean killOnTimeOut) throws IOException {
		synchronized (this) {
			// get all actually running worker processes
			Set<IWorkerProcess> ws = WorkerProcess.allWorkers();
			if (exclude!=null) {
				Iterator<IWorkerProcess> it = ws.iterator();
				while (it.hasNext()) { if (exclude.equals(it.next().getComponentName())) it.remove(); }
			}
			List<WorkerNotifier> t = new ArrayList<WorkerNotifier>(ws.size());
			for (IWorkerProcess wp : ws) {
				t.add(new WorkerNotifier(args, wp, timeout,killOnTimeOut));
			}
			ApplicationThreadPool.instance().execute(true,t);
			// collect errors
			Map<String, Throwable> c = new HashMap<String, Throwable>();
			for (WorkerNotifier n : t) {
				if (n.getError() != null) {
					c.put(n.getId(), n.getError());
				}
			}
			if (c.size() > 0)
				throw new MessageDistributionException("Error during message distribution", c);
		}
	}

	public void broadcastMessage(Map<String, Serializable> args, long timeout,boolean killOnTimeOut) throws IOException {
		this._broadcastMessage(args, timeout, null, killOnTimeOut);
	}
	
	public void broadcastMessage(Map<String, Serializable> args, long timeout) throws IOException {
		this._broadcastMessage(args, timeout, null, false);
	}
	
	private final static Logger logger = Logger.getLogger(HomeLayoutResource.class.getName());
}
